
#ifndef JV_WRITER_H
#define JV_WRITER_H 

#include <CGAL/basic.h>
#include <CGAL/Inverse_index.h>
#include <iostream>
#include <list>

 CGAL_BEGIN_NAMESPACE

/** This class enables writing jvx format (which is xml) according to
 * the jvx.dtd available at
 * http://www.javaview.de/guide/formats/Format_Jvx.html . The output
 * may be for a std::ostream or a CGAL::Javaview_stream. Metadata
 * (i.e. the title) and geometry attributes (i.e. name, colors, ...) 
 * are handled. Functions to write the core geometry of cgal objects
 * are not members of this class; these functions are global and some
 * of them defined with the operator <<. Geometries must be defined
 * between the write_header() and write_footer() functions.
 */

/* The template Out_stream must have an insert operator << for
   std::string (orchar*), int, double, ... that writes in ascii
   format. */
//other write_members can be designed for color, vector fields....

template <class Out_stream>
class Javaview_writer {
 public:
  Javaview_writer(Out_stream& out_stream) : 
    out(&out_stream), current_color(WHITE), show_geom(true) {}
   
  void write_header();
  void write_footer();
  void set_geometry_name(std::string name) {geometry_name = name; }
  void begin_geometry();
  void end_geometry();
  Out_stream* get_stream() { return out; }
  void set_title(std::string t_name) { title = t_name; }
  void set_color(Color new_color) { current_color = new_color; }
  Color get_color() { return current_color; }
  void write_color();
  void set_show_geom(bool b) { show_geom = b;}
  bool get_show_geom() { return show_geom; }

 protected:
  Out_stream* out;
  std::string geometry_name;
  std::string title;
  Color current_color;
  bool show_geom;
  //may be added: parameters for meta data

};

/** Needed to begin any output */
template <class Out_stream>
void Javaview_writer<Out_stream>::
write_header() {
  *out << "<?xml version=\"1.0\" encoding=\"ISO-8859-1\" standalone=\"no\"?>\n"
       << "<!DOCTYPE jvx-model SYSTEM \"http://www.javaview.de/rsrc/jvx.dtd\">\n"
       << "<jvx-model>\n"
       << "   <meta generator=\"CGAL Javaview_writer\"/>\n";
  if (!title.empty()) *out << "   <title> " << title << " </title>\n";
  *out << "   <geometries>\n";
}
  
/** Needed to and any output */
template <class Out_stream>
void Javaview_writer<Out_stream>::
write_footer() {
  *out << "   </geometries>\n"
       << "</jvx-model>\n";
}

/** Needed to begin any geometry group */
template <class Out_stream>
void Javaview_writer<Out_stream>::
begin_geometry() {
  *out << "      <geometry name=\"" << geometry_name << "\">\n";
}

template <class Out_stream>
void Javaview_writer<Out_stream>::
end_geometry() {
  *out << "      </geometry>\n";
}

template <class Out_stream>
void Javaview_writer<Out_stream>::
write_color() {
  *out << "               <color>" 
       << (int) current_color.r() << " " 
       << (int) current_color.g() << " "
       << (int) current_color.b() << "</color>\n";
}

//////////////END CLASS JAVAVIEW_WRITER////////////////////


///////GLOBAL FUNCTIONS///////////////////
/** \defgroup Global_writing_fcts Global functions for writing elements of geometries
 *
 * These global functions enable to write the elements of a
 * geometry. A geometry is composed of a pointSet or pointSet and
 * lineSet or pointSet and faceSet. In addition vectorField may be
 * defined for each of these sets. Other attributes are such as colors
 * may be added. To define a geometry these functions must be called
 * between a begin_geometry() and a end_geometry() of the class
 * Javaview_writer.
 */

/** \ingroup Global_writing_fcts */
/** Point_iter is an iterator on CGAL::Point_3d. The current color
 *  used for all points.
 */

template <class Point_iter, class Out_stream>
void pointSet_javaview_writer(Javaview_writer<Out_stream>& javaview_writer, 
			      Point_iter begin, Point_iter end)
{
  assert( end != begin);
  
  *javaview_writer.get_stream() << "         <pointSet dim=\"3\" point=";
  if (javaview_writer.get_show_geom())  *javaview_writer.get_stream() << "\"show\">\n";
  else *javaview_writer.get_stream() << "\"hide\">\n";
  *javaview_writer.get_stream()  << "            <points>\n";

  for(; begin != end; ++begin) 
    *javaview_writer.get_stream() << "               <p>"
				 << ::CGAL::to_double( begin->x()) << " "
				 << ::CGAL::to_double( begin->y()) << " "
				 << ::CGAL::to_double( begin->z()) 
				 << "</p>\n";
  javaview_writer.write_color();
  *javaview_writer.get_stream() << "            </points>\n"
				<< "         </pointSet>\n";
}

/** \ingroup Global_writing_fcts */
/** Point_iter is an iterator on CGAL::Point_3. A polyline connecting
 *  the given points is drawn with the current color.
 */
template <class Point_iter, class Out_stream>
  void polyline_javaview_writer(Javaview_writer<Out_stream>& javaview_writer, 
				Point_iter begin, Point_iter end)
{
  Point_iter second_point = begin;
  ++second_point;
  assert( (begin!=end) && (second_point!=end) );

  //do not show points of the polyline
  bool show_current_status = javaview_writer.get_show_geom();
  javaview_writer.set_show_geom(false);
  pointSet_javaview_writer(javaview_writer, begin, end);
  javaview_writer.set_show_geom(show_current_status);

  *javaview_writer.get_stream() << "         <lineSet line=\"show\">\n"
			       << "            <lines>\n";
  *javaview_writer.get_stream() << "               <l>";
  for(int i = 0; begin!=end; begin++ ) *javaview_writer.get_stream()  << i++ << " ";
  *javaview_writer.get_stream() << "</l>\n";
  
  javaview_writer.write_color();
  *javaview_writer.get_stream()	<< "            </lines>\n"
				<< "         </lineSet>\n";
}


/** \ingroup Global_writing_fcts */
/** A polyhedral surface (i.e. a CGAL::Polyhedron_3 or a class derived
 *  from it) is drawn with the current color.
 */
template <class Polyhedron, class Out_stream>
  void polyhedron_javaview_writer (Javaview_writer<Out_stream>& javaview_writer,
				   const Polyhedron& P)
{
  typedef typename Polyhedron::Vertex_const_iterator                  VCI;
  typedef typename Polyhedron::Facet_const_iterator                   FCI;
  typedef typename Polyhedron::Halfedge_around_facet_const_circulator HFCC;
  
  //pointSet from vertices
  *javaview_writer.get_stream()  << "         <pointSet dim=\"3\" point=\"hide\">\n"
				 << "            <points>\n";

  for( VCI vi = P.vertices_begin(); vi != P.vertices_end(); ++vi) 
    *javaview_writer.get_stream() << "               <p>"
				  << ::CGAL::to_double( vi->point().x()) << " "
				  << ::CGAL::to_double( vi->point().y()) << " "
				  << ::CGAL::to_double( vi->point().z()) << " "
				  << "</p>\n";

  *javaview_writer.get_stream() << "            </points>\n"
				<< "         </pointSet>\n";

  //faceSet
  *javaview_writer.get_stream() << "         <faceSet face=\"show\" edge=\"hide\">\n"
				<< "            <faces>\n";
  
  typedef Inverse_index< VCI> Index;
  Index index( P.vertices_begin(), P.vertices_end());
   
  for( FCI fi = P.facets_begin(); fi != P.facets_end(); ++fi) {
    HFCC hc = fi->facet_begin();
    HFCC hc_end = hc;
    std::size_t n = circulator_size( hc);
    CGAL_assertion( n >= 3);
    *javaview_writer.get_stream()  << "               <f>"; 
    do {
      *javaview_writer.get_stream() << " " << int(index[ VCI(hc->vertex())]);
      ++hc;
    } while( hc != hc_end);
    *javaview_writer.get_stream() << "</f>\n";
  }
  
 javaview_writer.write_color();
 *javaview_writer.get_stream() << "            </faces>\n"
			       << "         </faceSet>\n";
}

//////////////OPERATORS <<  ////////////////////////

/** \defgroup Insert_operators Insert operators to output a given geometry 
 * 
 * These operators enable to output a complete geometry, you cannot
 * add additional items.
 */

/** \ingroup Insert_operators */
/** output a point as a single point pointSet geometry */
template < class Kernel, class Out_stream >
Javaview_writer<Out_stream>&
operator<<(Javaview_writer<Out_stream>& jvw,
	   Point_3<Kernel>& point)
{
  jvw.begin_geometry();
  *jvw.get_stream() << "         <pointSet dim=\"3\" point=\"show\">\n"
		    << "            <points>\n"
		    << "               <p>"
		    << ::CGAL::to_double( point.x()) << " "
		    << ::CGAL::to_double( point.y()) << " "
		    << ::CGAL::to_double( point.z()) 
		    << "</p>\n";
  jvw.write_color();
  *jvw.get_stream() << "            </points>\n"
		    << "         </pointSet>\n";
  jvw.end_geometry();
  return jvw;
}

/** \ingroup Insert_operators */
/** Output a segment as a single line lineSet geometry */
template < class Kernel, class Out_stream >
Javaview_writer<Out_stream>&
operator<<(Javaview_writer<Out_stream>& jvw,
	   Segment_3<Kernel>& segment)
{
  jvw.begin_geometry();
  //pointSet
  std::list<typename Kernel::Point_3> points;
  points.push_back(segment.source());
  points.push_back(segment.target());
  //do not show points of the segment
  bool show_current_status = jvw.get_show_geom();
  jvw.set_show_geom(false);
  pointSet_javaview_writer(jvw, points.begin(), points.end());
  jvw.set_show_geom(show_current_status);
  //lineSet
  *jvw.get_stream() << "         <lineSet line=\"show\">\n"
		    << "            <lines>\n"
		    << "               <l>0 1</l>\n";
  jvw.write_color();
  *jvw.get_stream() << "            </lines>\n"
		    << "         </lineSet>\n";
  jvw.end_geometry();
  return jvw;
}

/** \ingroup Insert_operators */
/** Output a triangle as a single pointSet and faceSet geometry */
template < class Kernel, class Out_stream >
Javaview_writer<Out_stream>&
operator<<(Javaview_writer<Out_stream>& jvw,
	   Triangle_3<Kernel>& triangle)
{
  std::list<Point_3<Kernel> > points;
  for(int i=0; i<3; i++) points.push_back(triangle[i]);
  
  jvw.begin_geometry();
  //do not show points (vertices) of the triangle
  bool show_current_status = jvw.get_show_geom();
  jvw.set_show_geom(false);
  pointSet_javaview_writer(jvw, points.begin(), points.end());
  jvw.set_show_geom(show_current_status);
 
  //faceSet
  *jvw.get_stream() << "         <faceSet face=\"show\" edge=\"hide\">\n"
				<< "            <faces>\n"
				<< "               <f>0 1 2</f>\n"
    				<< "            </faces>\n"
				<< "         </faceSet>\n";
  jvw.end_geometry();
  return jvw;
}

/** \ingroup Insert_operators */
/** Output a single geometry consisting of a polyhedron. */
template < class Polyhedron, class Out_stream >
Javaview_writer<Out_stream>&
operator<<(Javaview_writer<Out_stream>& jvw,
	   Polyhedron& P)
{
  jvw.begin_geometry();
  polyhedron_javaview_writer(jvw, P);
  jvw.end_geometry();
  return jvw;
}

 CGAL_END_NAMESPACE

#endif // JV_WRITER_H //
// EOF //

